%%=========================================================================
%   Code for clustered multitask diffusion LMS    
%           --- Algorithm and convergence model  
%
% Reference paper:
%  J. Chen, C. Richard and A. H. Sayed,
%  Multitask diffusion LMS over network. 
%  IEEE ICASSP, 2014.
%
%  Contact:   Dr.Jie.Chen@ieee.org
%             http://www.jie-chen.com
%
% (Please note that the algorithm runs very fast, but the convergence model
% evaluatoin is very slow if the network/the size of the parameter vector is
% large, due to the kroncker product.)

clc, clear

% Node number
N = 15;
% Realization averaging times
RLZ = 1;      

% Length of the regressor
L = 2;

% Stepize
mu = 0.03;

% Regularization parameer
lambda  = 0.1;

% run time 
T = 1000;

% Topology generation
gene_topo


% If the following line is active, there will be no cluster structre:
%ClsMask = eye(N);        


% Glboal cost decomposition matrix:
%C = diag(1./sum(CM.*ClsMask))*CM.*ClsMask;  
%C = ones(N);
C = eye(N);
        

% Regularization weight

% Lapalcian matrix
P = CM.*(1-ClsMask);
idx = (sum(P,2)>0);
P(idx,:) = diag(1./sum(P(idx,:),2))*P(idx,:);
%P=eye(N);

% Combination matrix
A = diag(1./sum(CM.*ClsMask))*CM.*ClsMask; 
%A= eye(N);

% Signal power
%var_x = (rand(N,1)*0.4-0.2) + 1;
%var_z = 0.2*((rand(N,1)*0.4-0.2) + 2);
load varxz  


% system coefficients
ws0 =  [0.5, -0.4]';
%ws1  =  kron(ones(6,1),ws0+ 0.05*randn(L,1));
%ws2  =  kron(ones(4,1),ws0+ 0.05*randn(L,1));
%ws3  =  kron(ones(5,1),ws0+ 0.05*randn(L,1));
%ws = [ws1;ws2;ws3];

ws1 = (ws0+0.02*randn(L,1))*ones(1,6);
ws2 = (ws0+0.02*randn(L,1))*ones(1,4);
ws3 = (ws0+0.02*randn(L,1))*ones(1,5);
ws = [ws1, ws2, ws3];
wsv = reshape(ws,[],1);

% We can load a stored weights:
load w

% Node coefficient

Wav = zeros(L,N, T);
MSD = zeros(T,1);
MSDav = MSD;

% Simulation
X = zeros(L,N);
dN = zeros(N,1);

Qt=(diag(sum((P+P')))-(P+P'))/2;

for rlz = 1 : RLZ
rlz
W = zeros(L,N, T);
U = zeros(L,N, 1);


for t = 2 : T
    

     for n = 1 : N
         X(:,n) = sqrt(var_x(n))*randn(L,1);
         dN(n) = X(:,n)'*ws(:,n) + sqrt(var_z(n))*randn;
     end
     
     for n = 1 : N
            idx = find(P(n,:)~=0);
            if length(idx) >=1
                 regterms =  W(:,n,t-1)*ones(1,length(idx))-W(:,idx,t-1);
                 weightP = abs(P(n,idx)+P(idx,n)')/2;
                 regterm = sum(regterms*diag(weightP),2);
             else
                 regterm =0;
            end
            U(:,n) = W(:,n,t-1) + mu*(dN(n)-W(:,n,t-1)'*X(:,n))*X(:,n) - mu*lambda*regterm;
     end
     
     for n = 1 : N
         idx = (A(n,:)>0);
         W(:,n,t) = sum(U(:,idx)*diag(A(n,idx)),2);    
     end
     MSD(t) = norm(W(:,:,t)-ws,'fro')^2/N ;
end
        Wav = Wav+W;
        MSDav = MSDav + MSD;
end

Wav = Wav/RLZ;
MSDav = MSDav/RLZ;


figure, plot(10*log10(MSDav)), hold on

figure

plot(reshape(Wav,[],t)','b'), hold on
%plot((ws*ones(1,T))','r')


%%   Analytical model
% You can comment this part if you donot need it. This part is very slow

% Note that our matrix A is the transpose of that in the paper, so we write:
A=A';


AI  = kron(A,eye(L));
Q = kron(diag(sum((P+P')))-(P+P'),eye(L))/2;
%Q = eye(N*L) - kron(P,eye(L));
CI = kron(C,eye(L));
Rb = kron(var_x,eye(L))';
HR = bdiag(Rb*CI);

Evt = zeros(L*N,T);
Evt(:,1) = -wsv;

%Ev_inf = - mu *lambda*inv(AI'*(eye(N*L)-mu*(HR+lambda*Q)) - eye(N*L))*AI'*Q*ws;

B= AI'*(eye(N*L)-mu*(HR+lambda*Q));
K = kron(B' , B');
G = AI'*CI'*bdiag(kron(var_x.*var_z,eye(L))')*CI*AI;
g = reshape(G',[],1);
sig_msd = inv(eye((N*L)^2)-K)*reshape(eye(N*L),[],1)/N;
Sig_msd = reshape(sig_msd,N*L,N*L);
r = AI'*Q*wsv;

Ev_inf = -mu*lambda*inv(eye(N*L)-B)*r;

f_inf = mu^2*lambda^2*r'*Sig_msd*r - 2*mu*lambda*(r'*Sig_msd*B*Ev_inf);

MSD_inf = mu^2*trace(G*Sig_msd) + f_inf;
10*log10(MSD_inf)

MSDt = zeros(T,1);
MSDt(1) = wsv'*wsv/N;

sigI = reshape(eye((N*L)),[],1);


v0 = -wsv;
MecK = (eye((N*L)^2)-K);
v0v0 = kron(v0',v0');
rr = kron(r',r');
Kt = eye(size(K));
sKBv = zeros(size(sigI))';%kron((B*v0)',r')*(K-eye(size(K)));
for t = 2 : T
     Evt(:,t) = B*Evt(:,t-1) - mu*lambda*AI'*Q*wsv;
     
    
   %  MSDt(t) = MSDt(t-1) + 1/N*(mu^2*g'*(Kt*sigI) + mu^2*lambda^2*rr*(Kt*sigI) -2* mu*lambda*kron((B*Evt(:,t-1))',r')*(Kt*sigI)-v0v0*MecK*sigI);
      Sig = reshape(Kt*sigI, N*L,N*L);
      MSDt(t) = MSDt(t-1) + 1/N*(mu^2*trace(G*Sig) + mu^2*lambda^2*r'*Sig*r -2* mu*lambda*(sKBv*sigI + r'*B*Evt(:,t-1))-v0v0*MecK*sigI);
     Kt = K*Kt;
     MecK = MecK*K;
     sKBv = sKBv*K + kron((B*Evt(:,t-1))',r')*(K-eye(size(K)));
end

plot((Evt + wsv*ones(1,T))','m','linewidth',1)


%norm(- mu *lambda*inv(AI'*(eye(N*L)-mu*(HR+lambda*Q)) - eye(N*L))*AI'*Q*ws)

%norm(Wav(:,T)-ws)


clr = 'b';
figure, hold on

plot(ones(size(MSDav))*10*log10(MSD_inf),[clr,'-.'],'linewidth',2)
plot(10*log10(MSDt),[clr],'linewidth',2)
plot(10*log10(MSDav),[clr,'--'],'linewidth',2)

xlabel('Iteration $n$','fontsize',14,'interpreter', 'latex')
ylabel('MSD','fontsize',14,'interpreter', 'latex')
